unit CastleLineMath;

{$mode ObjFPC}{$H+}

interface

uses
  Math,CastleVectors;
  {
    Минимальное расстояние от точки до отрезка. С расчетом точки проекции PP.
    0 - точка проецируется на отрезок: Distance = расст.от P до отрезка SP1,SP2. PP - проекция точки P на отрезок SP1,SP2
    1 - точка P лежит "левее"  SP1: Distance = расст.от P до SP1. PP=SP1
    2 - точка P лежит "правее" SP2: Distance = расст.от P до SP2. PP=SP2
  }
  function PSDistanceEx(const P,SP1,SP2:TVector2;var Distance:Single; var PP:TVector2):integer;

  {
   Intersection of segments A1:A2, B1:B2
  returns
    1 if there is one intersection point "CrossPoint"
    0 if chunks ar on parallel lines
  -1 if there are no intersection points
  }
  function CrossingSegments(const a1,a2,b1,b2:TVector2; var CrossPoint:TVector2):shortint;

  {
  угол между отрезком v1:v2 и положительным направлением оси ОХ
  }
  function AngleOX(v1,v2:TVector2):single;

implementation


function PSDistanceEx(const P,SP1,SP2:TVector2;var Distance:Single; var PP:TVector2):integer;
var
  dX21,dY21,dX1,dY1,dX2,dY2,aSide,d,L2:single;
begin
  dX21:=SP2.X-SP1.X;
  dY21:=SP2.Y-SP1.Y;
  dX1:=P.X-SP1.X;
  dY1:=P.Y-SP1.Y;
  //если угол между векторами (SP1->SP2) и (SP1->P)  > +-90 градусов
  //=> P лежит "левее" точки SP1 => мин.Dist = расст.до точки SP1
  aSide:=dX21*dX1+dY21*dY1;
  if aSide <= 0 then
  begin
    if aSide = 0 then
      Result:=0
    else
      Result:=1;
    Distance:=sqrt(dX1*dX1+dY1*dY1);
    PP:=SP1;
  end
  else
  begin
    dX2:=P.X-SP2.X;
    dY2:=P.Y-SP2.Y;
    //если угол между векторами (SP2->SP1) и (SP2->P)  > +-90 градусов
    //=> P лежит "правее" точки SP2 => мин.Dist = расст.до точки SP2
    aSide:=dX21*dX2+dY21*dY2;
    if aSide >= 0 then  //= ((-dX21)*dX2+(-dY21)*dY2 < 0)
    begin
      if aSide = 0 then
        Result:=0
      else
        Result:=2;
      Distance:=sqrt(dX2*dX2+dY2*dY2);
      PP:=SP2;
    end
    else
    //оба угла <= 90гр. => точка проецируется на отрезок (SP1-SP2)
    //=> мин.расст = расстоянию до прямой SP1-SP2
    begin
      d:=dY21*dX1-dX21*dY1;       //ненормированное расстояние до отрезка со знаком
      L2:=dX21*dX21+dY21*dY21;    //квадрат длины отрезка
      Distance:=abs(d)/sqrt(L2);  //расстояние
      d:=d/L2;
      PP.X:=P.X-dY21*d;
      PP.Y:=P.Y+dX21*d;
      Result:=0;
    end;
  end;
end;

function AngleOX(v1,v2:TVector2):single;
var dx,dy:single;
begin
dx:=v2.x-v1.x;
dy:=v1.y-v2.y;
if(dx>0)and(dy>=0) then result:=arctan(dy/dx)else
  if (dx>0)and(dy<0) then result:=arctan(dy/dx)+2*pi else
    if dx<0 then result:=arctan(dy/dx)+pi;
end;

{
 Intersection of segments A1:A2, B1:B2
returns
  1 if there is one intersection point "CrossPoint"
  0 if chunks ar on parallel lines
-1 if there are no intersection points
}
function CrossingSegments(const a1,a2,b1,b2:TVector2; var CrossPoint:TVector2):shortint;
 const eps = 0.000001;
 var d,da,db,ta,tb: real;
begin
  CrossPoint:=Vector2(0,0);
  d :=(a1.x-a2.x)*(b2.y-b1.y) - (a1.y-a2.y)*(b2.x-b1.x);
  da:=(a1.x-b1.x)*(b2.y-b1.y) - (a1.y-b1.y)*(b2.x-b1.x);
  db:=(a1.x-a2.x)*(a1.y-b1.y) - (a1.y-a2.y)*(a1.x-b1.x);
  if (abs(d)<eps) then Result := 0
  else begin
   ta:=da/d;
   tb:=db/d;
   if (0<=ta)and(ta<=1)and(0<=tb)and(tb<=1)then begin
    CrossPoint:=Vector2(a1.x+ta*(a2.x-a1.x),a1.y+ta*(a2.y-a1.y));
    Result := 1
   end else Result := -1;
  end;
end;

end.

