unit CastleLines;

interface

uses CastleLinesGizmos,CastleClassUtils,CastleTransform, // для редактора
  CastleLineMath,Classes,CastleColors,CastleVectors,CastleScene,X3DNodes,CastleLinesBase,CastleTriangulate;


type
  TVector2DinamicArray = array of TVector2;
  TJoinMode=(jmBevel);

  TCastleLines = class (TCastleLinesBase)
    private
      FRects:TVector2DinamicArray;
      {соеденения линий в вершинах}
      FJoin:TVector2DinamicArray;
      {Тип соеденения}
      FJoinMode:TJoinMode;
      {треугольники для заполнения}
      FFill:TVector2DinamicArray;
      {Для определения нужно ли перестраивать сцену при изменении Scale}
      SignScale:Single;
      {перерисовать прямоугольники}
      procedure SetRects;
      {заполнить}
      procedure SetFill;
      {для триангуляции многоугольника для заполнения}
      procedure  AddFill(const Tri: TVector3Integer);
      function GetPoint(Index: integer): TVector3;
      {Сортирует вершины треугольника в порядок обхода по часовой стрелке}
      procedure DoGoodTriangle(var A,B,C:TVector2);

    protected
      function BuildScene: TX3DRootNode; override;
      procedure ChangedTransform; override;
    public
      procedure Clear; override;
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;

      {для сохранения в редакторе--------------------------------------------------}
      procedure Update(const SecondsPassed: Single; var RemoveMe: TRemoveType); override;
    published

  end;

implementation

uses SysUtils,CastleComponentSerialize;



procedure TCastleLines.Update(const SecondsPassed: Single; var RemoveMe: TRemoveType);
var i,tb:integer;
begin
 inherited;
 // для редактора
 tb:=0;
 for i:=0 to Self.Count-1 do begin
  if Self.Items[i].ClassName='TVisualizeTransformSelected.TGizmoScene' then begin
   LinesGizmos.Hide(Self);
   Exit;
  end;
  if Self.Items[i].ClassName= 'TDebugTransformBox.TInternalScene' then tb:=tb+1;
 end;
 if tb>=1 then begin
  LinesGizmos.Show(Self);
 end else begin
  LinesGizmos.Hide(Self);
 end;

end;


constructor TCastleLines.Create(AOwner: TComponent);
begin
  inherited;

  SignScale:=1;
  Setup2D;
  SetLength(FRects,0);
  SetLength(FJoin,0);
  SetLength(FFill,0);
  FJoinMode:=jmBevel;
  AddPoint(Vector2(10,10));
  AddPoint(Vector2(90,90));
  AddPoint(Vector2(170,10));
  AddPoint(Vector2(170,90));
  AddPoint(Vector2(10,10));
  ReLoad;


  // для редактора
{  Gizmos:=TCastleLinesGizmos.Create(Self);
  Gizmos.SetTransient;
  Add(Gizmos);
  Gizmos.ReLoad;
  Gizmos.Hide;}


  //----

end;

destructor TCastleLines.Destroy;
begin
  SetLength(FRects,0);
  SetLength(FJoin,0);
  SetLength(FFill,0);
  inherited;
end;

procedure TCastleLines.Clear;
begin
  inherited;
  SetLength(FRects,0);
  SetLength(FJoin,0);
  SetLength(FFill,0);
end;

function TCastleLines.BuildScene: TX3DRootNode;
var
  {}
  TransformNode:TTransformNode;
  {}
  GRects:TTriangleSet2DNode;
  ShapeRects:TShapeNode;
  MaterialRects:TMaterialNode;
  {}
  GJoin:TTriangleSet2DNode;
  ShapeJoin:TShapeNode;
  {}
  GFill:TTriangleSet2DNode;
  ShapeFill:TShapeNode;
  MaterialFill:TMaterialNode;


begin
  SetRects;
  SetFill;
  TransformNode:=TTransformNode.Create;
  // прямоугольники
  GRects:=TTriangleSet2DNode.CreateWithShape(ShapeRects);
  GRects.SetVertices(FRects);
  MaterialRects := TMaterialNode.Create;
  MaterialRects.EmissiveColor := WhiteRGB;
  MaterialRects.Transparency:=0.5;
  ShapeRects.Material := MaterialRects;
  // промежутки между прямоугольниками
  GJoin:=TTriangleSet2DNode.CreateWithShape(ShapeJoin);
  GJoin.SetVertices(FJoin);
  ShapeJoin.Material := MaterialRects;


  TransformNode.AddChildren(ShapeJoin);
  TransformNode.AddChildren(ShapeRects);


  // заполнение
  if LineType=ltPolygon then begin
   GFill:=TTriangleSet2DNode.CreateWithShape(ShapeFill);
   GFill.SetVertices(FFill);
   MaterialFill := TMaterialNode.Create;
   MaterialFill.EmissiveColor := YellowRGB;
//   MaterialFill.Transparency:=0.5;
   ShapeFill.Material := MaterialFill;
   TransformNode.AddChildren(ShapeFill);
  end;




  Result := TX3DRootNode.Create;
  Result.AddChildren(TransformNode);



end;

procedure TCastleLines.ChangedTransform;
var Sc:single;
begin
  inherited;
  Sc:=Scale.X*Scale.Y;
  if ((Sc<0)and(SignScale>0))or((Sc>0)and(SignScale<0)) then begin
   SignScale:=Sc;
   Load(BuildScene,true);
  end;
end;

procedure TCastleLines.DoGoodTriangle(var A,B,C:TVector2);
var ABx,ABy,ACx,ACy,P,Sc:Single;
    D:TVector2;
begin
 ABx:=B.x-A.x;
 ABy:=B.y-A.y;
 ACx:=A.x-C.x;
 ACy:=A.y-C.y;

 P:=(ABx * ACy - ACx * ABy);
 Sc:=Scale.X*Scale.Y;
 // при отрицательных Scale порядок обхода меняется
 if ((P>0)and(Sc>0))or((P<0)and(Sc<0))  then begin
  D:=B; B:=C; C:=D;
 end;
end;

procedure TCastleLines.SetRects;
var
    cross:TVector2;
    i,pi1,pi2:integer;
    d,dx,dy:single;
    angle:single;
begin
   if LineWidth<=0 then begin
    SetLength(FRects,0);
    SetLength(FJoin,0);
    Exit;
   end;
   SetLength(FRects,(CountPoints-1)*6);
   // заполняем вершины прямоугольников
   for pi1:=0 to CountPoints-2 do begin
    pi2:=pi1+1;
      d:=LineWidth/2;
      if Points[pi1].x=Points[pi2].x then begin // если вертикальная линия
       FRects[pi1*6+0]:=Vector2(Points[pi1].x-d,Points[pi1].y);
       FRects[pi1*6+1]:=Vector2(Points[pi1].x+d,Points[pi1].y);
       FRects[pi1*6+2]:=Vector2(Points[pi2].x+d,Points[pi2].y);
       FRects[pi1*6+3]:=Vector2(Points[pi2].x-d,Points[pi2].y);
       FRects[pi1*6+4]:=FRects[pi1*6+0];
       FRects[pi1*6+5]:=FRects[pi1*6+2];
      end else begin
       angle:= AngleOX(Points[pi1],Points[pi2]);
       dx:=(d*sin(angle)); //проекции d на оси ОХ и ОY
       dy:=(d*cos(angle));
       FRects[pi1*6+0]:=Vector2((Points[pi1].x-dx),(Points[pi1].y-dy));
       FRects[pi1*6+1]:=Vector2((Points[pi1].x+dx),(Points[pi1].y+dy));
       FRects[pi1*6+2]:=Vector2((Points[pi2].x+dx),(Points[pi2].y+dy));
       FRects[pi1*6+3]:=Vector2((Points[pi2].x-dx),(Points[pi2].y-dy));
       FRects[pi1*6+4]:=FRects[pi1*6+0];
       FRects[pi1*6+5]:=FRects[pi1*6+2];
      end;
   //   DoGoodTriangle(FRects[pi1*6+0],FRects[pi1*6+1],FRects[pi1*6+2]);
   //   DoGoodTriangle(FRects[pi1*6+3],FRects[pi1*6+4],FRects[pi1*6+5]);
   end;

   // --------------------------Прямоугольники на стыках не должны пересекаться
   for i:=1 to CountPoints-1 do begin
      pi1:=i-1;
      pi2:=i;
      if i=CountPoints-1 then begin        // необходимо обработать последний и первй прямоугольник
        if LineType<>ltOpen then begin     // если это не открытая линия
         pi1:=i-1;
         pi2:=0;
        end else Continue;
      end;
      if CrossingSegments(FRects[pi1*6+1],FRects[pi1*6+2],FRects[pi2*6+1],FRects[pi2*6+2],cross) = 1 then begin
         FRects[pi1*6+2]:=cross;
         FRects[pi2*6+1]:=cross;
         FRects[pi1*6+5]:=FRects[pi1*6+2];
       end else
       if CrossingSegments(FRects[pi1*6+0],FRects[pi1*6+3],FRects[pi2*6+0],FRects[pi2*6+3],cross) = 1 then begin
         FRects[pi1*6+3]:=cross;
         FRects[pi2*6+0]:=cross;
         FRects[pi2*6+4]:=FRects[pi2*6+0];
       end else
       if CrossingSegments(FRects[pi1*6+1],FRects[pi1*6+2],FRects[pi2*6+0],FRects[pi2*6+3],cross) = 1 then begin
         FRects[pi1*6+2]:=cross;
         FRects[pi2*6+0]:=cross;
         FRects[pi1*6+5]:=FRects[pi1*6+2];
         FRects[pi2*6+4]:=FRects[pi2*6+0];
       end else
       if CrossingSegments(FRects[pi1*6+0],FRects[pi1*6+3],FRects[pi2*6+1],FRects[pi2*6+2],cross) = 1 then begin
         FRects[pi1*6+3]:=cross;
         FRects[pi2*6+1]:=cross;
       end;
   end;

   //----------------------------------------- Cоеденения между прямоугольниками
    if FJoinMode=jmBevel then begin

     if LineType<>ltOpen then SetLength(FJoin,(CountPoints-1)*3) else SetLength(FJoin,(CountPoints-2)*3);
     for i:=0 to CountPoints-2 do begin
      pi1:=i;
      pi2:=i+1;
      if i=CountPoints-2 then begin        // необходимо обработать последний и первый прямоугольник
        if LineType<>ltOpen then begin     // если это не открытая линия
          pi1:=i;
          pi2:=0;
        end else Continue;
      end;
     // заполнить промежутки между прямоугольниками
      if ((FRects[pi1*6+3].X=FRects[pi2*6+0].X) and (FRects[pi1*6+3].Y=FRects[pi2*6+0].Y)) or
         ((FRects[pi1*6+2].X=FRects[pi2*6+0].X) and (FRects[pi1*6+2].Y=FRects[pi2*6+0].Y)) then begin
       FJoin[pi1*3+0]:=FRects[pi1*6+2];
       FJoin[pi1*3+1]:=FRects[pi1*6+3];
       FJoin[pi1*3+2]:=FRects[pi2*6+1];
      end else
      if ((FRects[pi1*6+2].X=FRects[pi2*6+1].X) and (FRects[pi1*6+2].Y=FRects[pi2*6+1].Y)) or
         ((FRects[pi1*6+3].X=FRects[pi2*6+1].X) and (FRects[pi1*6+3].Y=FRects[pi2*6+1].Y)) then begin
       FJoin[pi1*3+0]:=FRects[pi1*6+3];
       FJoin[pi1*3+1]:=FRects[pi1*6+2];
       FJoin[pi1*3+2]:=FRects[pi2*6+0];
      end else begin
       FJoin[pi1*3+0]:=FRects[pi1*6+0];
       FJoin[pi1*3+1]:=FRects[pi1*6+0];
       FJoin[pi1*3+2]:=FRects[pi1*6+0];
      end;
     end;
    end;
   //----- вершины треугольников должны быть в порядке обхода по часавой стрелке
   for i:=0 to CountPoints-2 do begin
    if (FJoinMode=jmBevel)and(i*3<Length(FJoin)) then DoGoodTriangle(FJoin[i*3+0],FJoin[i*3+1],FJoin[i*3+2]);
    DoGoodTriangle(FRects[i*6+0],FRects[i*6+1],FRects[i*6+2]);
    DoGoodTriangle(FRects[i*6+3],FRects[i*6+4],FRects[i*6+5]);
   end;
end;


procedure  TCastleLines.AddFill(const Tri: TVector3Integer);
var cnt:integer;
begin
   cnt:=Length(FFill);
   SetLength(FFill,cnt+3);
   FFill[cnt+0]:=Points[Tri.X];
   FFill[cnt+1]:=Points[Tri.Y];
   FFill[cnt+2]:=Points[Tri.Z];
end;
function TCastleLines.GetPoint(Index: integer): TVector3;
begin
   Result:=Vector3(Points[index].X,Points[index].Y,0);
end;

procedure TCastleLines.SetFill;
var i,j:integer;
    isCross:boolean;
    cross:TVector2;
begin
  SetLength(FFill,0);
  if LineType<>ltPolygon then Exit;
  // проверка на самопересечение
  // пересечения надо искаь через отрезок
  isCross:=false;
  for i:=0 to CountPoints-3 do begin
   for j:=i+2 to CountPoints-3 do begin
     if CrossingSegments(Points[i],Points[i+1],Points[j],Points[j+1],Cross)=1 then isCross:=true;
     if isCross then break;
   end;
    if isCross then break;
  end;
  // проверка последнего отрезка
  for j:=1 to CountPoints-4 do begin
    if CrossingSegments(Points[CountPoints-2],Points[CountPoints-1],Points[j],Points[j+1],Cross)=1 then isCross:=true;
    if isCross then break;
  end;


  // триангуляция
  if not isCross then TriangulateFace(nil,CountPoints,@GetPoint,@AddFill,0);
end;



initialization

RegisterSerializableComponent(TCastleLines, 'Line Transform');

end.
