unit CastleLinesBase;

interface

uses
   Classes,CastleScene,X3DNodes,CastleVectors,CastleClassUtils;
type
  TVector2DinamicArray = array of TVector2;
  {обычная линия, замкнутая линия, многоуголтник}
  TLineType=(ltOpen,ltClose,ltPolygon);

  TCastleLinesBase = class (TCastleScene)
  private
    {Количество точек}
    FCountPoints:integer;
    {точки}
    FPoints:TVector2DinamicArray;
    {Ширина линии}
    FLineWidth:Single;
    {Тип линии (обычная линия, замкнутая линия, многоуголтник)}
    FLineType:TLineType;
    {получить точку линии}
    function  GetPoint(index:integer):TVector2;
  protected
    {Построение сцены}
    function BuildScene: TX3DRootNode; virtual abstract;
    {установить ширину линии}
    procedure SetLineWidth(wl:Single); virtual;
    {изменить точку линии}
    procedure SetPoint(index:integer; newPoint:TVector2); virtual;
    {Установить тип линии}
    procedure SetLineType(NewLineType:TLineType); virtual;
    {привести линиую к соответствие типу}
    procedure CheckLineType; virtual;
  public
    {Добавить точку}
    procedure AddPoint(Point:TVector2); virtual;
    {вставить точку на место index}
    procedure InsertPoint(index:integer;point:TVector2); virtual;
    {удалить точку}
    procedure DelPoint(index:integer); virtual;
    {удалить все точки}
    procedure Clear; virtual;
    {для сохранения в редакторе}
    procedure CustomSerialization(const SerializationProcess: TSerializationProcess); override;
    {нарисовать заново}
    procedure ReLoad;
    {constructor}
    constructor Create(AOwner: TComponent); override;
    {destructor}
    destructor Destroy; override;

    {точки}
    property Points[i:integer]:TVector2 read GetPoint write SetPoint;
    {количество точек}
    property CountPoints :integer read FCountPoints;
  published
    {ширина линии}
    property LineWidth:Single read FLineWidth write SetLineWidth;
    {Тип линии (обычная линия, замкнутая линия, многоуголтник)}
    property LineType:TLineType read FLineType write SetLineType;
  end;


implementation

uses SysUtils;

{constructor}
constructor TCastleLinesBase.Create(AOwner: TComponent);
begin
  inherited;
  FCountPoints:=0;
  SetLength(FPoints,0);
  FLineWidth:=10;
  FLineType:=ltOpen;
end;

{destructor}
destructor TCastleLinesBase.Destroy;
begin
 SetLength(FPoints,0);
 FCountPoints:=0;
 inherited;
end;

{нарисовать заново}
procedure TCastleLinesBase.ReLoad;
begin
 Load(BuildScene,true);
end;

{для сохранения в редакторе}
procedure TCastleLinesBase.CustomSerialization(const SerializationProcess: TSerializationProcess);
var i:integer;
begin
  SerializationProcess.ReadWriteInteger('FCountPoints',FCountPoints,true);
  SetLength(FPoints,FCountPoints);
  for i:=0 to FCountPoints-1 do begin
    SerializationProcess.ReadWriteVector('FPoints'+inttostr(i), FPoints[i],vector2(0,0), true);
  end;
  CheckLineType;
  inherited;
  ReLoad;
end;

{удалить все точки}
procedure TCastleLinesBase.Clear;
begin
  FCountPoints:=0;
  SetLength(FPoints,0);
end;

{удалить точку}
procedure TCastleLinesBase.DelPoint(index:integer);
var i:integer;
begin
  if (index<0) or (index>=FCountPoints) or  (FCountPoints<1) then Exit;

  if (FLineType<>ltOpen)and(index=FCountPoints-1) then index:=0;

  for i:=index to FCountPoints-2 do begin
      fpoints[i]:=fpoints[i+1];
  end;
  SetLength(FPoints,FCountPoints-1);
  FCountPoints:=FCountPoints-1;


  if (FLineType<>ltOpen)and(index=0)and(FCountPoints>2) then begin
   SetLength(FPoints,FCountPoints-1);
   FCountPoints:=FCountPoints-1;
  end;
  CheckLineType;
end;

{вставить точку на место index}
procedure TCastleLinesBase.InsertPoint(index:integer;point:TVector2);
var i:integer;
begin
  if index>=FCountPoints then begin
   AddPoint(point);
   Exit;
  end;
  if index<0 then index:=0;

  SetLength(FPoints,FCountPoints+1);
  for i:=FCountPoints downto index+1 do begin
      FPoints[i]:=FPoints[i-1];
  end;
  FCountPoints:=FCountPoints+1;

  fpoints[index]:=point;

end;

{Добавить точку}
procedure TCastleLinesBase.AddPoint(Point:TVector2);
begin
  if (FLineType<>ltOpen)and(FCountPoints>3) then begin
   InsertPoint(FCountPoints-1,Point);
  end else begin
   SetLength(FPoints,FCountPoints+1);
   FPoints[FCountPoints]:=Point;
   FCountPoints:=FCountPoints+1;
   if (FLineType<>ltOpen)and(FCountPoints=3) then CheckLineType;
  end;

end;

{установить ширину линии}
procedure TCastleLinesBase.SetLineWidth(wl:Single);
begin
  if FLineWidth=wl then Exit;
  FLineWidth:=wl;
  ReLoad;
end;

{изменить точку линии}
procedure TCastleLinesBase.SetPoint(index:integer; newPoint:TVector2);
begin
  FPoints[index]:=newPoint;
  if (FLineType<>ltOpen)and(FCountPoints>2) then begin
   if index=0 then FPoints[FCountPoints-1]:=newPoint;
   if index=FCountPoints-1 then FPoints[0]:=newPoint;
  end;
end;

{Установить тип линии}
procedure TCastleLinesBase.SetLineType(NewLineType:TLineType);
begin
 FLineType:=NewLineType;
 CheckLineType;
 Reload;
end;

{получить точку линии}
function  TCastleLinesBase.GetPoint(index:integer):TVector2;
begin
  Result:=FPoints[index];
end;

{привести линиую к соответствие типу}
procedure TCastleLinesBase.CheckLineType;
begin
 if FCountPoints<3 then Exit;
 if FLineType=ltOpen then Exit;
 if (FCountPoints=3)and(FPoints[0].X=FPoints[2].X)and(FPoints[0].Y=FPoints[2].Y) then begin
  DelPoint(2);
  Exit;
 end;
 if (FCountPoints>2)and((FPoints[0].X<>FPoints[FCountPoints-1].X)or(FPoints[0].Y<>FPoints[FCountPoints-1].Y)) then begin
  SetLength(FPoints,FCountPoints+1);
  FPoints[FCountPoints]:=FPoints[0];
  FCountPoints:=FCountPoints+1;
  Exit;
 end;



end;


end.

