unit CastleLinesGizmos;

interface

uses CastleLinesBase,CastleTransform,CastleTriangles,CastleRenderOptions,CastleLineMath,
     Classes,CastleColors,CastleVectors,CastleScene,X3DNodes,Math,CastleKeysMouse;

const
  {Растояние до точки при котором точка выделяется}
  DISTANCE_SELECTED_POINT=8;
  {Растояние до линии при котором рисуется точка для вставки}
  DISTANCE_INSERT_POINT=4;
  {Ширина линии для редактирования}
  LINE_WIDTH_EDITOR=2;
  {Расстояние до границе}
  BORDER_DISTANCE=50;
type
  TCastleLinesGizmos = class (TCastleScene)
    private

      {выбранная вершина (если ничего не выбрано, то =-1)}
      FSelectPoint:integer;
      {для вставки новой точки в редакторе}
      FInsertPoint:TVector2;
      {индекс для вставки новой точки в редакторе}
      FInsertPointIndex:integer;
      {включен механизм перетаскивания}
      isMovePoint:boolean;
      {Линия которую редактирует}
      Line:TCastleLinesBase;
    protected
      function LocalRayCollision(const RayOrigin, RayDirection: TVector3;const TrianglesToIgnoreFunc: TTriangleIgnoreFunc): TRayCollision; override;
    public
      function PointingDeviceMove(const Pick: TRayCollisionNode; const Distance: Single): Boolean;override;
      function PointingDevicePress(const Pick: TRayCollisionNode; const Distance: Single): Boolean; override;
      function PointingDeviceRelease(const Pick: TRayCollisionNode; const Distance: Single; const CancelAction: Boolean): Boolean; override;
      {Для отслеживания нажатия правой кнопки мыши}
      function Press(const E: TInputPressRelease): boolean; override;
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      function BuildScene: TX3DRootNode;
      procedure ReLoad;
      {спрятать и убрать все выбранные точки}
      Procedure Hide(ALine:TCastleLinesBase);
      {отобразить и перерисовать}
      Procedure Show(ALine:TCastleLinesBase);


    published


  end;

  var LinesGizmos:TCastleLinesGizmos;

implementation


uses SysUtils;


constructor TCastleLinesGizmos.Create(AOwner: TComponent);
begin
  inherited;
  Line:=nil;
  Setup2D;
  Self.SetTransient;
  Hide(nil);

  ListenPressRelease:=true;   // что бы работали события мыши
  RenderOptions.PointSize:=DISTANCE_SELECTED_POINT;
  RenderOptions.LineWidth:=LINE_WIDTH_EDITOR;
  RenderLayer := rlFront;


  // надо понять что это за параметры!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  InternalExistsOnlyInMeaningfulParents := true;
  Collides := false;
  Pickable := true;
  CastShadows := false;
  ExcludeFromStatistics := true;
  InternalExcludeFromParentBoundingVolume := true;

  //------------------------------------------------

end;

destructor TCastleLinesGizmos.Destroy;
begin
  inherited;
end;

procedure TCastleLinesGizmos.ReLoad;
begin
  Load(BuildScene,true);
end;

Procedure TCastleLinesGizmos.Hide(ALine:TCastleLinesBase);
begin
  if ALine<>Line then Exit;
  if Self.Exists=false then Exit;
  FSelectPoint:=-1;
  FInsertPointIndex:=-1;
  FInsertPoint:=TVector2.Zero;
  isMovePoint:=false;
  Exists:=false;

  if Line<>nil then Line.Remove(Self);
  Line:=nil;
end;

Procedure TCastleLinesGizmos.Show(ALine:TCastleLinesBase);
begin
  if ALine=Line then Exit;
  if Self.Exists then Exit;
  Line:=ALine;
  if Line<>nil then Line.Add(Self);

  Exists:=true;
  ReLoad;
end;


function TCastleLinesGizmos.LocalRayCollision(const RayOrigin, RayDirection: TVector3;
  const TrianglesToIgnoreFunc: TTriangleIgnoreFunc): TRayCollision;
begin
 Result := inherited;
 if Result <> nil then Result.Distance := 0;
end;

function TCastleLinesGizmos.PointingDevicePress(const Pick: TRayCollisionNode; const Distance: Single): Boolean;
begin
  Result:=inherited;
  if Result then Exit;
   if FSelectPoint>=0 then begin
    isMovePoint:=True;
    Result:=True;
   end;
end;

function TCastleLinesGizmos.Press(const E: TInputPressRelease): boolean;
begin
  Result:=inherited;
  if Result then Exit;
  if (E.IsMouseButton(buttonRight)and(FSelectPoint>=0)and(isMovePoint=false))and(Line.CountPoints>2) then begin  // удаление точки
   Line.DelPoint(FSelectPoint);
   FSelectPoint:=-1;
   Line.ReLoad;
   Reload;
  end;
end;

function TCastleLinesGizmos.PointingDeviceMove(const Pick: TRayCollisionNode; const Distance: Single): boolean;
var i:integer;
    point:TVector3;
    OldSelectPoint:integer;
    OldInsertPoint:TVector2;
    PP:TVector2;
    dist:Single;

begin
   Result:=inherited;
   if Result then Exit;
   Point:=OutsideToLocal(Pick.Point);
 {   // учет положения
   Point:=(Pick.Point-Translation);
   // учет поворота
   if Rotation.W<>0 then Point:=Center+RotatePointAroundAxisRad(-Rotation.W,Point-Center,Vector3(Rotation.X,Rotation.Y,Rotation.Z));
   // учет масштаба
   Point:=Vector3(Point.X-Center.X,Point.Y-Center.Y,Point.Z-Center.Z)*Vector3(1/Scale.X,1/Scale.Y,1/Scale.Z)+Center;
  }
   if isMovePoint=False then begin
     OldSelectPoint:=FSelectPoint;
     // выделение точки
     FSelectPoint:=-1;
     for i:=Line.CountPoints -1 downto 0  do begin
      if PointsDistance(Point.XY,Line.Points[i])<DISTANCE_SELECTED_POINT then begin
       FSelectPoint:=i;
       Result:=true;
       break;
      end;
     end;

     // вставка точки на линию
         FInsertPointIndex:=-1;
         OldInsertPoint:=FInsertPoint;
         FInsertPoint:=TVector2.Zero;
         PP:=TVector2.Zero;
         dist:=DISTANCE_INSERT_POINT+1;
         if FSelectPoint=-1 then begin
          for i:=Line.CountPoints-1 downto 1 do begin
           if PSDistanceEx(Point.XY,Line.Points[i],Line.Points[i-1],dist,PP)=0 then begin
            if dist<DISTANCE_INSERT_POINT then begin
             FInsertPointIndex:=i;
             FInsertPoint:=PP;
             break;
            end; {if dist<DISTANCE_INSERT_POINT}
           end; {if PSDistanceEx....}
          end; {for}
         end; {if fselectPoint=-1}
         // end вставка точки на линию


     if (OldSelectPoint<>FSelectPoint)or(FInsertPoint.X<>OldInsertPoint.X)or(FInsertPoint.Y<>OldInsertPoint.Y) then Load(BuildScene,true);


   end else begin // Механизм перемещения
    Result:=true;
    Line.Points[FSelectPoint]:=Point.XY;

    Line.ReLoad;
    Reload;
   end;

end;


function TCastleLinesGizmos.PointingDeviceRelease(const Pick: TRayCollisionNode; const Distance: Single; const CancelAction: Boolean): Boolean;
var
     point:TVector3;
begin
 Result:=inherited;
 if Result then Exit;

 if isMovePoint then begin  // конец перемещения точки
  isMovePoint:=False;
  Result:=True
 end else
 if FInsertPointIndex>=0 then begin // вставить точку
   if FSelectPoint=-1 then begin
     Line.InsertPoint(FInsertPointIndex,FInsertPoint);
     FSelectPoint:=FInsertPointIndex;
     FInsertPointIndex:=-1;
     FInsertPoint:=TVector2.Zero;
     Line.ReLoad;
     Reload;
   end;
 end else begin // добавление точки
  if FSelectPoint=-1 then begin
    Point:=OutsideToLocal(Pick.Point);
    Line.AddPoint(Point.XY);
    Line.ReLoad;
    Reload;
  end;
 end;

end;

function TCastleLinesGizmos.BuildScene: TX3DRootNode;
var
  {}
  TransformNode:TTransformNode;
  {}
  GPoints:TPolypoint2DNode;
  ShapePoints:TShapeNode;
  MaterialPoints:TMaterialNode;
  {}
  GInsertPoint:TPolypoint2DNode;
  ShapeInsertPoint:TShapeNode;
  MaterialInsertPoint:TMaterialNode;
  {}
  GSelectPoint:TPolypoint2DNode;
  ShapeSelectPoint:TShapeNode;
  MaterialSelectPoint:TMaterialNode;
  {}
  GLines:TPolyline2DNode;
  ShapeLines:TShapeNode;
  MaterialLines:TMaterialNode;
  {}
  GBorder:TPolyline2DNode;
  ShapeBorder:TShapeNode;
  MaterialBorder:TMaterialNode;

  Points:TVector2DinamicArray;
  i:integer;
  MaxXY,MinXY:TVector2;
begin
  Result := TX3DRootNode.Create;
  if Line.CountPoints<=0 then Exit;
  SetLength(Points,Line.CountPoints);
  MaxXY:=line.Points[0];
  MinXY:=line.Points[0];
  for i:=0 to Line.CountPoints-1 do begin
   points[i]:=line.Points[i];
   if Line.CountPoints>1 then begin
     if MaxXY.X<points[i].X then MaxXY.X:=points[i].X;
     if MaxXY.Y<points[i].Y then MaxXY.Y:=points[i].Y;
     if MinXY.X>points[i].X then MinXY.X:=points[i].X;
     if MinXY.Y>points[i].Y then MinXY.Y:=points[i].Y;
   end;
  end;
  MaxXY:=MaxXY+Vector2(BORDER_DISTANCE,BORDER_DISTANCE);
  MinXY:=MinXY-Vector2(BORDER_DISTANCE,BORDER_DISTANCE);

  TransformNode:=TTransformNode.Create;

  //граница
  GBorder:=TPolyline2DNode.CreateWithShape(ShapeBorder);
  GBorder.SetLineSegments([MinXY,Vector2(MinXY.X,MaxXY.Y),MaxXY,Vector2(MaxXY.X,MinXY.Y),MinXY]);
  MaterialBorder:=TMaterialNode.Create;
  MaterialBorder.EmissiveColor := RedRGB;
  ShapeBorder.Material := MaterialBorder;


  //линии
  GLines:=TPolyline2DNode.CreateWithShape(ShapeLines);
  GLines.SetLineSegments(Points);
  MaterialLines := TMaterialNode.Create;
  MaterialLines.EmissiveColor := RedRGB;
  ShapeLines.Material := MaterialLines;
  // точки вершин
  GPoints:=TPolypoint2DNode.CreateWithShape(ShapePoints);
  GPoints.SetPoint(Points);
  MaterialPoints := TMaterialNode.Create;
  MaterialPoints.EmissiveColor := RedRGB;
  ShapePoints.Material := MaterialPoints;
  // выделенная точка
  if FSelectPoint>=0 then begin;
   GSelectPoint:=TPolypoint2DNode.CreateWithShape(ShapeSelectPoint);
   GSelectPoint.SetPoint(Points[FSelectPoint]);
   MaterialSelectPoint:=TMaterialNode.Create;
   MaterialSelectPoint.EmissiveColor := GrayRGB;
   ShapeSelectPoint.Material:=MaterialSelectPoint;
   TransformNode.AddChildren(ShapeSelectPoint);
  end;
  // точка для вставки на линию
  if FInsertPointIndex>=0 then begin
   GInsertPoint:=TPolypoint2DNode.CreateWithShape(ShapeInsertPoint);
   GInsertPoint.SetPoint(FInsertPoint);
   MaterialInsertPoint:=TMaterialNode.Create;
   MaterialInsertPoint.EmissiveColor := BlueRGB;
   ShapeInsertPoint.Material:=MaterialInsertPoint;
   TransformNode.AddChildren(ShapeInsertPoint);
  end;

  TransformNode.AddChildren(ShapeBorder);
  TransformNode.AddChildren(ShapePoints);
  TransformNode.AddChildren(ShapeLines);

  Result.AddChildren(TransformNode);

  SetLength(Points,0);
end;


initialization
  LinesGizmos:=TCastleLinesGizmos.Create(nil);

finalization
   LinesGizmos.Free;






end.

